/*
 * Barcode finder thingy.
 * Copyright (C) 1997 By Simon T. Kirby.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "settings.h"
#include "grab.h"
#include "videodev.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <Imlib.h>

/*+- Typedefs -+*/

typedef unsigned int dword;
typedef unsigned short word;
typedef unsigned char byte; 

/*+- Defines -+*/

#define X_STEP 24			/* X stride */
#define Y_STEP 24			/* Y stride */
#define D_STEP (X_STEP + Y_STEP)	/* Diagonal stride */

#define MAX_DEVIATION_FACTOR 0.49	/* Maximum pixel deviation per point */
#define MAX_PIXELS_PER_POINT 16		/* Maximum number of pixels per point */


#define EAN_BITS_PER_NUMBER 7	/* Number of points (bits) per number */
#define EAN_RLES_PER_NUMBER 4	/* Number of bars/spaces per number */
#define EAN_RLE_MAX 4		/* Maximum number of modules in a single bar */

//#define UPC_A_POINTS 95		/* Number of points (bits) total */

#define UPC_A_NUMBERS 12	/* Number of digits */
#define UPC_A_RLES (UPC_A_NUMBERS * EAN_RLES_PER_NUMBER + 3 + 3 + 5)

#define UPC_E_NUMBERS 7		/* Number of digits */
#define UPC_E_RLES (UPC_E_NUMBERS * EAN_RLES_PER_NUMBER + 3 + 2)

#define MIN_RLES UPC_E_RLES	/* Minimum number of RLEs */
#define MAX_RLES UPC_A_RLES	/* Maximum number of RLEs */

#define CHANGE_TO_ONE_FOR_FUN 0

#undef VERBOSE
#undef LIGHT_VERBOSE

/*+- Globals -+*/

#if 0
unsigned char ean_bits_a_str[][EAN_BITS_PER_NUMBER + 1] = {
	{ 0,0,1,1,0,1,0, },
	{ 0,1,1,0,0,1,0, },
	{ 0,1,0,0,1,1,0, },
	{ 1,1,1,1,0,1,0, },
	{ 1,0,0,0,1,1,0, },
	{ 1,1,0,0,0,1,0, },
	{ 1,0,1,1,1,1,0, },
	{ 1,1,1,0,1,1,0, },
	{ 1,1,0,1,1,1,0, },
	{ 0,0,1,0,1,1,0, },
};

unsigned char ean_bits_a_inversed_str[][EAN_BITS_PER_NUMBER + 1] = {
	{ 1,1,1,0,0,1,0, },
	{ 1,1,0,0,1,1,0, },
	{ 1,1,0,1,1,0,0, },
	{ 1,0,0,0,0,1,0, },
	{ 1,0,1,1,1,0,0, },
	{ 1,0,0,1,1,1,0, },
	{ 1,0,1,0,0,0,0, },
	{ 1,0,0,0,1,0,0, },
	{ 1,0,0,1,0,0,0, },
	{ 1,1,1,0,1,0,0, },
};

#else

unsigned char ean_digits_rles[][EAN_RLES_PER_NUMBER] = {
/*
 * Even parity (UPC right side)
 *        B S B S          UPCleft UPCright
 */	{ 3,2,1,1 },	/* 0001101 1110010 0 */
	{ 2,2,2,1 },    /* 0011001 1100110 1 */
	{ 2,1,2,2 },    /* 0010011 1101100 2 */
	{ 1,4,1,1 },    /* 0111101 1000010 3 */
	{ 1,1,3,2 },    /* 0100011 1011100 4 */
	{ 1,2,3,1 },    /* 0110001 1001110 5 */
	{ 1,1,1,4 },    /* 0101111 1010000 6 */
	{ 1,3,1,2 },    /* 0111011 1000100 7 */
	{ 1,2,1,3 },    /* 0110111 1001000 8 */
	{ 3,1,1,2 },    /* 0001011 1110100 9 */
/*        S B S B
 * Odd parity (UPC left side)
 *
11132111411131232113211321111111113211322221121332113112111	UPC-A
111 3121 1411 3112 1123 1231 1141 111111111			UPC-E
    rev  fwd  fwd  rev  fwd  rev
111312114113112112312311141111111111
 */
};

#endif

static int global_threshold;

int debug, have_dga = 0;
int x11_is_remote = 1;  
char v4l_conf[] = "";   

char tvnorm[] = TV_NORM;
char input[] = INPUT_SOURCE;

char *device;

int width = RESOLUTION_X;
int height = RESOLUTION_Y;

int linelength = 0;

/*+- Functions -+*/

static void image_rerender();

static int barcode_upc_a_check_not_guards(unsigned char *rles){
	int point,number,i,d;
	char result[UPC_A_NUMBERS + 1];

	point = 3;
	d = 0;
	for (number = 0;number < 6;number++){
		for (i = 0;i < 10;i++)
			if (memcmp(rles + point,ean_digits_rles[i],EAN_RLES_PER_NUMBER) == 0){
				result[d++] = i + '0';
				break;
			}
		if (i == 10){
#ifdef LIGHT_VERBOSE
			fprintf(stderr,"Unable to decode at rle %i (%u%u%u%u).\n",
                                point,rles[point],rles[point+1],rles[point+2],rles[point+3]);
#endif
			return 0;
		}
		point+= EAN_RLES_PER_NUMBER;
	}
	point+= 5;
	for (number = 0;number < 6;number++){
		for (i = 0;i < 10;i++)
			if (memcmp(rles + point,ean_digits_rles[i],EAN_RLES_PER_NUMBER) == 0){
				result[d++] = i + '0';
				break;
			}
		if (i == 10){
#ifdef LIGHT_VERBOSE
			fprintf(stderr,"Unable to decode at rle %i (%u%u%u%u).\n",
                                point,rles[point],rles[point+1],rles[point+2],rles[point+3]);
#endif
			return 0;
		}
		point+= EAN_RLES_PER_NUMBER;
	}
	result[d] = '\0';
	printf("---------> %s <--------",result);
	fflush(stdout);
{static int success_cnt = 0;fprintf(stderr,"\007 -- SUCCESS!!!!!! (%u) [Thresh: %u]\n",success_cnt++,global_threshold);}
	return 1;
}

/*
 * Validate a UPC-A style barcode.
 */
static int barcode_upc_a_check(unsigned char *rles){
	unsigned char rles_rev[UPC_A_RLES + 1];
	int i;

	/*
	 * Check guard bar widths
	 */

#ifdef VERBOSE
	fprintf(stderr,"Entering barcode_upc_a_check()...\n");
#endif
		
	if (rles[(UPC_A_RLES / 2) - 2] != 1)
		goto out;
	if (rles[(UPC_A_RLES / 2) - 1] != 1)
		goto out;
	if (rles[(UPC_A_RLES / 2) + 0] != 1)
		goto out;
	if (rles[(UPC_A_RLES / 2) + 1] != 1)
		goto out;
	if (rles[(UPC_A_RLES / 2) + 2] != 1)
		goto out;

#ifdef VERBOSE
	fprintf(stderr,"Passes middle 01010 check!  Woohoo!\n");
#endif

	if (rles[UPC_A_RLES - 1] != 1)
		goto out;
	if (rles[UPC_A_RLES - 2] != 1)
		goto out;
	if (rles[UPC_A_RLES - 3] != 1)
		goto out;

#ifdef VERBOSE
	fprintf(stderr,"Right 101 matched.  Passes end checks!\n");
#endif

	if (rles[0] != 1)
		goto out;
	if (rles[1] != 1)
		goto out;
	if (rles[2] != 1)
		goto out;

#ifdef VERBOSE
	fprintf(stderr,"Left 101 matched.\n");
#endif


	/*
	 * Guard bar checks pass, now check inner bars.
	 */

	if (barcode_upc_a_check_not_guards(rles))	/* Check forwards */
		return 1;
	for (i = 0;i < UPC_A_RLES;i++)
		rles_rev[i] = rles[UPC_A_RLES - i - 1];
	if (barcode_upc_a_check_not_guards(rles_rev))	/* Check backwards */
		return 1;
out:
	return 0;
}


static int barcode_upc_e_check_not_guards(unsigned char *rles){
	int point,number,i,d;
	char result[UPC_E_NUMBERS + 1];

#define LIGHT_VERBOSE
#if 0
fprintf(stderr,"Almost there... ");
for (d = 0;d < UPC_E_RLES;d++){
	fprintf(stderr,"%u",rles[d]);
}
fprintf(stderr,"\n");
#endif		
/*    rev  fwd  fwd  rev  fwd  rev */

	point = 3;
	d = 0;
	for (number = 0;number < 6;number++){
		for (i = 0;i < 10;i++)
			if ((number == 1) || (number == 2)  || (number == 3) /* BZZT */ || (number == 3)){
				if (memcmp(rles + point,ean_digits_rles[i],EAN_RLES_PER_NUMBER) == 0){
					result[d++] = i + '0';
					break;
				}
			} else {
				if ((rles[point + 3] == ean_digits_rles[i][0]) &&
				    (rles[point + 2] == ean_digits_rles[i][1]) &&
				    (rles[point + 1] == ean_digits_rles[i][2]) &&
				    (rles[point + 0] == ean_digits_rles[i][3])){
					result[d++] = i + '0';
					break;
				}
			}
		if (i == 10){
#ifdef LIGHT_VERBOSE
			fprintf(stderr,"Unable to decode at rle %i number %i (%u%u%u%u).\n",
                                point,number,rles[point],rles[point+1],rles[point+2],rles[point+3]);
#endif
			return 0;
		}
		point+= EAN_RLES_PER_NUMBER;
	}
	result[d] = '\0';
	printf("---------> %s <--------",result);
	fflush(stdout);
{static int success_cnt = 0;fprintf(stderr,"\007 -- SUCCESS!!!!!! (%u) [Thresh: %u]\n",success_cnt++,global_threshold);}
	return 1;
}

/*
 * Validate a UPC-E style barcode.
 */
static int barcode_upc_e_check(unsigned char *rles){
	unsigned char rles_rev[UPC_E_RLES + 1];
	int i;

	/*
	 * Check guard bar widths
	 */

	if (rles[UPC_E_RLES - 1] != 1)
		goto out;
	if (rles[UPC_E_RLES - 2] != 1)
		goto out;
	if (rles[UPC_E_RLES - 3] != 1)
		goto out;
	if (rles[UPC_E_RLES - 4] != 1)
		goto out;
	if (rles[UPC_E_RLES - 5] != 1)
		goto out;
	if (rles[UPC_E_RLES - 6] != 1)
		goto out;

	if (rles[0] != 1)
		goto out;
	if (rles[1] != 1)
		goto out;
	if (rles[2] != 1)
		goto out;

	/*
	 * Guard bar checks pass, now check inner bars.
	 */

	if (barcode_upc_e_check_not_guards(rles))	/* Check forwards */
		return 1;
	for (i = 0;i < UPC_E_RLES;i++)
		rles_rev[i] = rles[UPC_E_RLES - i - 1];
	if (barcode_upc_e_check_not_guards(rles_rev))	/* Check backwards */
		return 1;
out:
	return 0;
}

/*
 * Analyze a suspicious RLE chunk.
 */
static int barcode_rle_check(float ofs[],float len[],int value[],int num_points){
	float deviation;
	float length_by_value[2];
	int i,d,multiple;
	int retval;
	unsigned char rles[MAX_RLES + 1];

	/*
	 * Break down the RLE into binary data up to the maximum
	 * supported barcode length.
	 */
#if 0
	if (num_points < MIN_RLES)
		return 0;
#endif
	length_by_value[value[0]] = len[0];
	length_by_value[value[1]] = len[1];

//fprintf(stderr,"Initial deviation: %.3f %.3f\n",len[0],len[1]);
	d = 0;
	for (i = 0;i < num_points;i++){
		for (multiple = 1;multiple <= EAN_RLE_MAX;multiple++){
			deviation = fabs((length_by_value[value[i]] * multiple) - len[i]);
//fprintf(stderr,"Deviation for length[%u] (%.3f) for %.3f at multiple %u is %.3f.\n",i,len[i],length_by_value[value[i]],multiple,deviation);
			if (deviation <= length_by_value[value[i]] * MAX_DEVIATION_FACTOR){
//fprintf(stderr,"Found multiple.\n");
				length_by_value[value[i]] = (length_by_value[value[i]] * 3 + ((float)len[i] / (float)multiple)) / 4.;
//				length_by_value[value[i]] = (length_by_value[value[i]] + ((float)len[i] / (float)multiple)) / 2.;
//				length_by_value[value[i]] = ((float)len[i] / (float)multiple);

				rles[d++] = multiple;
//fprintf(stderr,"%u",multiple);

//				for (j = 0;j < multiple;j++)
//	fprintf(stderr,"%c",value[i] + '0');
//	fprintf(stderr,"\n");

				goto got_bit;
			}
			if (deviation < len[i] * multiple)
				continue;
//			fprintf(stderr,"High deviation of %.3f spotted in barcode_rle_check!\n",deviation);
			return 0;
		}
	got_bit:
		if (d == 3)
			if (rles[2] != 1)	/* 0 and 1 already have to be 1 */
				break;
		if (d >= MAX_RLES)
			break;
	}

	if (d < MIN_RLES)
		return 0;

	/* TODO: Implement barcode style plugin mechanism here. */

	if (d >= UPC_E_RLES)
		if ((retval = barcode_upc_e_check(rles)))
			return retval;
	if (d >= UPC_A_RLES)
		if ((retval = barcode_upc_a_check(rles)))
			return retval;
	return 0;
}


/*
 * Scan for a known barcode in the RLE data.
 */
static void barcode_analyze_scan(float ofs[],float len[],int value[],int num_points){
	float max_width;
	int i;

	/*
	 * In order for the entire barcode to fit, the most simple (short)
	 * barcode's longest initial bar is finite and easy to calculate.
	 *
	 * We're assuming the starting guard bars are the thinnest bars
	 * possible, which is valid for UPCs at least.
	 */
	max_width = (float)width / (6 * EAN_BITS_PER_NUMBER);

	for (i = 0;i < num_points - 4;i++){
		if (!value[i])	/* White */
			continue;
		if (len[i] > max_width)
			continue;

		barcode_rle_check(&ofs[i],&len[i],&value[i],num_points - i);
	}
}

inline float linear_interpolation_real_part(int value_left,int value_right,int threshold){
	/*
	 *             +
	 *  ---thresh---    or y-flipped  (in this case, return 0.5)
	 *  +
	 *
	 * Finding the percentage through which the threshold is found.
	 * Ratio of closeness to left (0) or right (1).
	 */

/*	if (value_left <= threshold)*/

	float real_part = (float)(threshold - value_left) / (float)(value_right - value_left);

	if (value_right == value_left)
		return 0.;
	if (real_part > 1){
//		fprintf(stderr,"linear_interpolation: real_part == %.3f (threshold: %u, value_left: %u, value_right: %u)\n",real_part,threshold,value_left,value_right);
//		abort();
		real_part = 1;
	}
	if (real_part < 0){
//		fprintf(stderr,"linear_interpolation: real_part == %.3f (threshold: %u, value_left: %u, value_right: %u)\n",real_part,threshold,value_left,value_right);
		real_part = 0;
	}
//return 0;
	return (float)(threshold - value_left) / (float)(value_right - value_left);
}

static int barcode_scan_threshold(byte *buf,char *rgb_buf,
                                  int xofs,int yofs,
                                  int xend,int yend,
                                  int xinc,int yinc,int threshold){
	float nonintegral_skew = 0.;
	float ofs[xend+yend];
	float len[xend+yend];
	int value[xend+yend];
	int index,i,j;
	int x,y;
	int pointofs,pointinc;
	int threshold_bool;
	byte last_pixel,pixel;

	global_threshold = threshold;

#define THRESHOLD_BOOLEAN(x) (x < threshold ? 1 : 0)

	pointofs = 0;
	pointinc = abs(xinc ? xinc : yinc);

	yinc*= xend;
	yofs*= xend;
	yend*= xend;

	last_pixel = buf[xofs + yofs];
	index = 0;
	value[index] = THRESHOLD_BOOLEAN(last_pixel);
	ofs[index] = 0;
	for (x = xofs,y = yofs;(x >= 0 && x < xend) && (y >= 0 && y < yend);x+= xinc,y+= yinc,pointofs+= pointinc){
		pixel = buf[x + y];
		threshold_bool = THRESHOLD_BOOLEAN(pixel);
		if (value[index] != threshold_bool){
			nonintegral_skew =
				linear_interpolation_real_part(pixel,last_pixel,threshold);
			ofs[index + 1] = (float)pointofs - nonintegral_skew;
			len[index] = ofs[index + 1] - ofs[index];
			index++;
			value[index] = threshold_bool;
		}
		last_pixel = pixel;
	}
	len[index] = (float)pointofs - nonintegral_skew - ofs[index];

	index++;

#if 0
	for (x = xofs,y = yofs;x < xend && y < yend;x+= xinc,y+= yinc){
		rgb_buf[(x + y) * 3 + 0] = 0;
		rgb_buf[(x + y) * 3 + 1] = 127;
		rgb_buf[(x + y) * 3 + 2] = 255;
	}
#endif
#if 1
	for (i = 0;i < index;i++){
		x = (int)ofs[i] * xinc + xofs;
		y = (int)ofs[i] * yinc + yofs;
		for (j = 0;j < len[i];j+= pointinc){
			rgb_buf[(x + y) * 3 + 0] = (value[i] ^ 1) * 0;
			rgb_buf[(x + y) * 3 + 1] = (value[i] ^ 1) * 63;
			rgb_buf[(x + y) * 3 + 2] = (value[i] ^ 1) * 255;
			x+= xinc;
			y+= yinc;
		}
	}
#endif

	/*
	 * See if we can find anything.
	 */
	barcode_analyze_scan(ofs,len,value,index);

	return threshold;
}

static void barcode_scan_thresholds(byte *buf,char *rgb_buf,
                                   int xofs,int yofs,
                                   int xend,int yend,
                                   int xinc,int yinc){

	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,250);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,240);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,230);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,225);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,220);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,210);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,200);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,190);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,180);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,170);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,160);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,150);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,130);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,110);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,95);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,80);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,65);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,50);
	barcode_scan_threshold(buf,rgb_buf,xofs,yofs,xend,yend,xinc,yinc,100);
}

static int barcode_scan(unsigned char *buf,unsigned char *rgb_buf,int w,int h)
{
	int x,y;

#if 0
	unsigned int threshold;
	int mean,min,max,v;

fprintf(stderr,"\rScan[%i]...",y);

		mean = 0; min = 255; max = 0;
		for (x = 0;x < w;x++){
			v = buf[l + x];
			mean+= v;
			if (v < min)
				min = v;
			if (v > max)
				max = v;
		}
		mean/= w;
		threshold = mean;

fprintf(stderr,"Threshold: %u Min: %u Max: %u  AvgMinMax: %u  Magic: %u\n",
        threshold,min,max,(min + max) / 2,(max + threshold) / 2 );
#endif

	for (y = 0;y < h;y+= Y_STEP)
		barcode_scan_thresholds(buf,rgb_buf,0,y,w,h,1,0);

	for (x = 0;x < w;x+= X_STEP)
		barcode_scan_thresholds(buf,rgb_buf,x,0,w,h,0,1);

	for (x = 0;x < w;x+= D_STEP)
		barcode_scan_thresholds(buf,rgb_buf,x,0,w,h,1,1);
	for (y = D_STEP;y < h;y+= D_STEP)
		barcode_scan_thresholds(buf,rgb_buf,0,y,w,h,1,1);

	for (x = D_STEP;x < w;x+= D_STEP)
		barcode_scan_thresholds(buf,rgb_buf,x,h - 1,w,h,1,-1);
	for (y = h - 1;y >= 0;y-= D_STEP)
		barcode_scan_thresholds(buf,rgb_buf,0,y,w,h,1,-1);

	return 1;
}

/*
 * Scan an RGB image for barcodes.
 *
 * Prepare the image for barcode scanning by filtering the red
 * channel out.
 */
static int barcode_scan_rgb(unsigned char *rgb,int w,int h)
{
	unsigned char *buf,*s;
	int retval;
	int i;

	s = buf = malloc(w * h);
	if (!buf)
		goto out_nomem;

	for (i = 0;i < w * h * 3;i+= 3)
		*s++ = rgb[i];

	for (i = 0;i < w * h * 3;i+= 3)
		rgb[i + 1] = rgb[i + 2] = rgb[i];

	retval = barcode_scan(buf,rgb,w,h);
	free(buf);
	return retval;

out_nomem:
	perror("malloc");
	return 0;
}

	Display *disp;
	Window win;
	ImlibData *id;
	ImlibImage *im;
	int w, h;
	Pixmap p, m;
	XSetWindowAttributes attr;

void image_rerender(){
	Imlib_changed_image(id,im);

	/* Re-render the Image to the new size */
	Imlib_render(id, im, w, h);
	/* Free the previous pixmap used for the window - note ONLY the pixmap is */
	/* freed - the mask is marked automatically to be freed along with the */
	/* pixmap. There is no need to free it as well - in fact it is advised you do */
	/* not. You must use the Imlib free function because of Imlib's caching. Do */
	/* not use any other free functions. You can use this function for Pixmaps */
	/* not created by Imlib - and it will just go free them as expected. */
	Imlib_free_pixmap(id, p);

	p = Imlib_move_image(id, im);
#if 0
	/* The mask will be 0 if the image has no transparency */
	m = Imlib_move_mask(id, im);
#endif
	/* Put the Image pixmap in the background of the window */
	XSetWindowBackgroundPixmap(disp, win, p);
	/* If there was a mask to the image, set the Image's mask to it */
#if 0
	if (m)
		XShapeCombineMask(disp, win, ShapeBounding,
				  0, 0, m, ShapeSet);
#endif
	/* Clear the window to update the background change */
	XClearWindow(disp, win);
#if 0
	/* Synchronise with the Xserver */
	XSync(disp, False);
#endif
}

void frame_grab(byte *buf){
	grabber_capture(buf,width * BYTES_PER_PIXEL,/* single frame */ 1);
}

int str_to_int(char *str,struct STRTAB *tab){
	int i;

	if (str[0] >= '0' && str[0] <= '9')
		return atoi(str);
	for (i = 0; tab[i].str != NULL; i++)
		if (strcasecmp(str,tab[i].str) == 0)
			return(tab[i].nr);
	return -1;
}

#define NUM_BUFFERS 1

int main(int argc, char **argv){
	int grab_mode = 0;
	ImlibImage faked_image;
	byte *buf[NUM_BUFFERS];
	byte *alpha[NUM_BUFFERS];
	int grabfd,i;

	/* Connect to the default Xserver */
	disp = XOpenDisplay(NULL);
	/* Immediately afterwards Intitialise Imlib */
	id = Imlib_init(disp);

	if (argc <= 1) {
		grab_mode = 1;

		device = "/dev/video";

		for (i = 0;i < NUM_BUFFERS;i++){
			buf[i] = (byte *)malloc(IMAGE_BUFFER_SIZE+1024);
			if (!buf[i]){
				perror("malloc()");
				exit(1);
			}
			memset(buf[i],0,IMAGE_BUFFER_SIZE);
		}

		for (i = 0;i < NUM_BUFFERS;i++){
			alpha[i] = (byte *)malloc(PIXELS+1024);
			if (!alpha[i]){
				perror("malloc()");
				exit(1);
			}
			memset(alpha[i],0,PIXELS);
		}

		grabfd = grabber_open(0,0,0,0,0);
		if ((!grabber->grab_setparams) ||
			 (!grabber->grab_capture)){	
			fprintf(stderr,"%s\ncapture not supported\n",grabber->name);
			exit(1);
		}

		/* Modify settings */
		{
			int norm = -1, in = -1;

			if (tvnorm && grabber->norms){
				norm = str_to_int(tvnorm,grabber->norms);
				if (norm == -1)
					fprintf(stderr,"Unknown TV norm: %s\n",tvnorm);
			}
			if (input && grabber->inputs){
				in = str_to_int(input,grabber->inputs);
				if (in == -1)
					fprintf(stderr,"Unknown input type: %s\n",input);
			}
			if (grabber->grab_input)
				grabber->grab_input(in,norm);
		}

		/* Set capture parameters */
		grabber_setparams(VIDEO_RGB24,&width,&height,&linelength,0);

		/* Brightness and stuff */
		/* color bright hue contrast*/
		grabber->grab_picture(32768,34816,32768 - 256,32768 - 10240 + 2048 + 4096 + 2048);

		im = &faked_image;

		im = Imlib_load_image(id, "640x480.png");

		im->rgb_width = w = width;
		im->rgb_height = h = height;
//		im->filename = NULL;
		im->rgb_data = buf[0];
//		im->alpha_data = alpha[0];

		/* Grab an inital frame */
		frame_grab(buf[0]);
	} else {
		/* Load the image specified as the first argument */
		im = Imlib_load_image(id, argv[1]);
		if (!im){
			fprintf(stderr,"Duhm, couldn't load image.\n");
			exit(0);
		}
		w = im->rgb_width;
		h = im->rgb_height;
	}
	/* Create a Window to display in */
	win = XCreateWindow(disp, DefaultRootWindow(disp), 0, 0, w, h, 0,
	                    id->x.depth, InputOutput, id->x.visual, 0, &attr);
	XSelectInput(disp, win, StructureNotifyMask);

	/* Render the original 24-bit Image data into a pixmap of size w * h */
	Imlib_render(id, im, w, h);
	/* Extract the Image and mask pixmaps from the Image */
	p = Imlib_move_image(id, im);
#if 0
	/* The mask will be 0 if the image has no transparency */
	m = Imlib_move_mask(id, im);
#endif
	/* Put the Image pixmap in the background of the window */
	XSetWindowBackgroundPixmap(disp, win, p);
	/* If there was a mask to the image, set the Image's mask to it */
#if 0
	if (m)
		XShapeCombineMask(disp, win, ShapeBounding, 0, 0, m,
				  ShapeSet);
#endif
	/* Actually display the window */
	XMapWindow(disp, win);
	/* Synchronise with the Xserver */
	XSync(disp, False);

	barcode_scan_rgb(im->rgb_data,w,h);

	if (grab_mode == 0)
		fprintf(stderr,"Barcode scanning finished.\n");

	if (grab_mode == 1){
		while (1){
			frame_grab(buf[0]);
			barcode_scan_rgb(im->rgb_data,w,h);
			image_rerender();
		}
	}

//exit(0);
	/* Event loop to handle resizes */
	for (;;) {
		XEvent ev;

		/* Sit and wait for an event to happen */
		XNextEvent(disp, &ev);
		if (ev.type == ConfigureNotify) {
			w = ev.xconfigure.width;
			h = ev.xconfigure.height;
			/* Re-render the Image to the new size */
			Imlib_render(id, im, w, h);
			/* Free the previous pixmap used for the window - note ONLY the pixmap is */
			/* freed - the mask is marked automatically to be freed along with the */
			/* pixmap. There is no need to free it as well - in fact it is advised you do */
			/* not. You must use the Imlib free function because of Imlib's caching. Do */
			/* not use any other free functions. You can use this function for Pixmaps */
			/* not created by Imlib - and it will just go free them as expected. */
			Imlib_free_pixmap(id, p);
			p = Imlib_move_image(id, im);
			/* The mask will be 0 if the image has no transparency */
			m = Imlib_move_mask(id, im);
			/* Put the Image pixmap in the background of the window */
			XSetWindowBackgroundPixmap(disp, win, p);
			/* If there was a mask to the image, set the Image's mask to it */
			if (m)
				XShapeCombineMask(disp, win, ShapeBounding,
						  0, 0, m, ShapeSet);
			/* Clear the window to update the background change */
			XClearWindow(disp, win);
			/* Synchronise with the Xserver */
			XSync(disp, False);
		}
	}
}
